/*****************************************************************************
 *   mcitest.c:  main C entry file for NXP LPC23xx/24xx Family Microprocessors
 *
 *   Copyright(C) 2006, NXP Semiconductor
 *   All rights reserved.
 *
 *   History
 *   2006.07.20  ver 1.00    Prelimnary version, first Release
 *
******************************************************************************/
#include "LPC23xx.h"                       /* LPC23xx/24xx definitions */
#include "type.h"
#include "irq.h"
#include "mci.h"
#include "dma.h"
#include "debug.h"


#define MCI_DMA_ENABLED	1

const char tbUniToASC[]={
197,96,96,170,96,178,175,96,96,96,
96,96,96,96,96,192,193,194,195,196,
197,198,199,200,201,202,203,204,205,206,
207,208,209,210,211,212,213,214,215,216,
217,218,219,220,221,222,223,224,225,226,
227,228,229,230,231,232,233,234,235,236,
237,238,239,240,241,242,243,244,245,246,
247,248,249,250,251,252,253,254,255,96,
96,96,96,186,73,179,191,};


extern volatile DWORD MCI_CardType;
extern volatile DWORD MCI_Block_End_Flag;
extern volatile BYTE *WriteBlock, *ReadBlock;

unsigned long MBR_sector_BIOS_PB; //  ()    ( 0) -  446 + 08h)
//***********************************************   BPB (BIOS Parameter Block) ************************************************
unsigned long BPB_SectSize; //     (  BIOS PB -  0Bh)[  512]
unsigned char BPB_ClustSize;//     (  BIOS PB -  ODh)
unsigned long BPB_ResSect;  //     FAT#1 (  BIOS PB -  0Eh)
unsigned char BPB_FatCnt;   // -  FAT   (  BIOS PB -  10h)[  2]
unsigned long BPB_RootSize; //  32-    (  BIOS PB -  11h)( FAT16   512)
unsigned long BPB_FatSize;  //     FAT (  BIOS PB -  16h)
//*************************************************************************************************************************************

//**********************************************   *******************************************************************
unsigned long FAT_begin; //  FAT (  )
unsigned long Root_dir_begin; //    (  )
unsigned long Data_begin; //    ( 2)  
//unsigned char SectorBuffer[512]; //    
unsigned long MMC_size;
unsigned long F_FileSize;
unsigned long F_StartCluster;
//******************************************************************\\

/********************************************************************/

unsigned char disk_initialize(void)
{
  unsigned char	err=0;
  DMA_Init();
  debug("DMA Init ok!\n");							 
  if ( MCI_Init() == TRUE )
	debug("MCI Init ok!\n");
  else 
	debug("MCI Init error!\n");  

  MCI_CardType = MCI_CardInit();

  if ( MCI_CardType == CARD_UNKNOWN )
	debug("error CARD_UNKNOWN\n");		/* fatal error */
  else { 
   if (MCI_CardType == MMC_CARD)
     debug("Find MMC_CARD\n");
   else	debug("Find SD_CARD\n");
  }

  if (MCI_Check_CID() != TRUE)
  {
	debug("Check_CID failed\n");		/* fatal error */
  }

  if ( MCI_Set_Address() != TRUE )
  {
	debug("Set_Address failed\n");		/* fatal error */
  }

  if ( MCI_Send_CSD() != TRUE )
  {
	debug("Send_CSD failed\n");		/* fatal error */
  }

  if ( MCI_Select_Card() != TRUE )
  {
	debug("Select_Card failed\n");		/* fatal error */
  }

  if ( MCI_CardType == SD_CARD )
  {
	MCI_Set_MCIClock( NORMAL_RATE );
	if (SD_Set_BusWidth( SD_4_BIT ) != TRUE )
	  debug("set 4 bit mode failed\n");	/* fatal error */
	else debug("set 4 bit mode\n");
  }

  if ( MCI_Set_BlockLen( BLOCK_LENGTH ) != TRUE )
	debug("Set_BlockLen failed\n");		/* fatal error */
  else 
  debug("Set_BlockLen o.k.\n");
 return err;
}
/*****************************************************************************/
void put_dump(unsigned int size)
{
	static int setbyte,cnt,tmp;
	cnt = 0;
	debug(" \n\000");
	for (setbyte=0;setbyte<size;setbyte++){
	 dump_chr(ReadBlock[setbyte]); 
	 cnt++;
	 if (cnt < 16) 
	   debug("-\000");
	 else { 
	   debug("   \000");
	 for (tmp = setbyte-16;tmp<setbyte;tmp++)
	 if (ReadBlock[tmp]>13)
	   putchar1(ReadBlock[tmp]);
	 else
	   putchar1('.');
	   debug(" \n\000");
	   cnt = 0;
	 }
   }
}

/*****************************************************************************/
unsigned char FAT_init(void) 
{
        unsigned char result;

    if ( MCI_Read_Block( 0 ) != TRUE )
	{
	  debug("Error MCI_Read_Block!\n");		/* Fatal error */
	  return result;
	}
       put_dump(512);
//        
        if (ReadBlock[510] != 0x55 || ReadBlock[511] != 0xAA)
         {
         debug("Card not formated!\n");
         }      

        BPB_SectSize = ReadBlock[0x0c];  //     (  BIOS PB)[  512]
        BPB_SectSize <<=8;  
        BPB_SectSize += ReadBlock[0x0b]; //
        
        BPB_ClustSize = ReadBlock[0x0D]; //     (  BIOS PB)
        
        BPB_ResSect = ReadBlock[ 0x0f];  //     FAT#1 (  BIOS PB) 
        BPB_ResSect <<=8;                   //
        BPB_ResSect += ReadBlock[ 0x0e]; //
       
        BPB_FatCnt = ReadBlock[0x10];    // -  FAT   (  BIOS PB)[  2]

        BPB_RootSize = ReadBlock[0x12];  //  32-    ( FAT16   512)
        BPB_RootSize <<=8;                  //                                                  ( FAT32   0)
        BPB_RootSize += ReadBlock[0x11]; //

        BPB_FatSize = ReadBlock[0x17];   //     FAT (  BIOS PB)
        BPB_FatSize  <<=8;                  //
        BPB_FatSize  += ReadBlock[0x16]; //
        

        MMC_size = ReadBlock[0x23];
        MMC_size <<=8;
        MMC_size+= ReadBlock[0x22];
        MMC_size <<=8;
        MMC_size+= ReadBlock[0x21];
        MMC_size <<=8;
        MMC_size+= ReadBlock[0x20];

        MMC_size=MMC_size/2048;
     	MBR_sector_BIOS_PB = 0;

        FAT_begin = MBR_sector_BIOS_PB + BPB_ResSect; //  FAT
        Root_dir_begin = FAT_begin + (BPB_FatSize * BPB_FatCnt); //   
        Data_begin = Root_dir_begin + (BPB_RootSize * 32 / 512); //    ( 2)

		debug("\nBPB_SectSize: %0\000",BPB_SectSize); 
		debug("\nBPB_ClustSize: %0\000",BPB_ClustSize);
		debug("\nBPB_RootSize: %0\000",BPB_RootSize);
		debug("\nBPB_FatSize: %0\000",BPB_FatSize);
		debug("\nMMC_size: %0\000",MMC_size);
		debug("\nFAT_begin: %0\000",FAT_begin);
		debug("\nRoot_dir_begin: %0\000",Root_dir_begin);
		debug("\nData_begin: %0\000",Data_begin);

        return 0;
}
/*****************************************************************************/
char UniToASC(unsigned char data1,unsigned char data2)
{
    int chr;
//	unsigned char res1;
	if (data2 == 0) return(data1);
	else{
	  chr = data2<<8;
	  chr+=data1;
	  	chr -= 1025;
		if (chr < 86)
		return (tbUniToASC[chr]);
	}
  return (data1);
}

/*****************************************************************************/
void FindFiles(void)
{
char F_LName[64];
char attrib,i,setL;
unsigned char record;
int root_sector;
static unsigned char setLL;
MCI_Read_Block(488);
	put_dump(512);
for (root_sector = 488; root_sector <= (Root_dir_begin+32);root_sector++){
	MCI_Read_Block(root_sector);
	debug("\nroot_sector:\n %0\000",root_sector);
	put_dump(512);
   record =0;
	setLL = 0;
    while(ReadBlock[record*32] !=0){
	if (ReadBlock[record*32] != 0xe5) { //    -   ,  
	  attrib = ReadBlock[(record*32)+11]; //  
	  if (attrib == 0x20 || attrib == 0x10 ){
	    F_StartCluster = ReadBlock[(record*32)+27]<<8;
	    F_StartCluster += ReadBlock[(record*32)+26];
		F_FileSize =  ReadBlock[(record*32)+30]<<16;
		F_FileSize +=  ReadBlock[(record*32)+29]<<8;
		F_FileSize += ReadBlock[(record*32)+28];
		}
	   else if (attrib == 0x0F) {
		 setL = ReadBlock[(record*32)];
		 setL &= 0x3f; 
		 if (setL > 0) setL--;
			 setL*=13;
			 F_LName[setL]    = UniToASC(ReadBlock[(record*32)+1],ReadBlock[(record*32)+2]);
			 F_LName[setL+1]  = UniToASC(ReadBlock[(record*32)+3],ReadBlock[(record*32)+4]);
			 F_LName[setL+2]  = UniToASC(ReadBlock[(record*32)+5],ReadBlock[(record*32)+6]);
			 F_LName[setL+3]  = UniToASC(ReadBlock[(record*32)+7],ReadBlock[(record*32)+8]);
			 F_LName[setL+4]  = UniToASC(ReadBlock[(record*32)+9],ReadBlock[(record*32)+10]);
			 F_LName[setL+5]  = UniToASC(ReadBlock[(record*32)+14],ReadBlock[(record*32)+15]);
			 F_LName[setL+6]  = UniToASC(ReadBlock[(record*32)+16],ReadBlock[(record*32)+17]);
			 F_LName[setL+7]  = UniToASC(ReadBlock[(record*32)+18],ReadBlock[(record*32)+19]);
			 F_LName[setL+8]  = UniToASC(ReadBlock[(record*32)+20],ReadBlock[(record*32)+21]);
			 F_LName[setL+9]  = UniToASC(ReadBlock[(record*32)+22],ReadBlock[(record*32)+23]);
			 F_LName[setL+10] = UniToASC(ReadBlock[(record*32)+24],ReadBlock[(record*32)+25]);
			 F_LName[setL+11] = UniToASC(ReadBlock[(record*32)+28],ReadBlock[(record*32)+29]);
			 F_LName[setL+12] = UniToASC(ReadBlock[(record*32)+30],ReadBlock[(record*32)+31]);
			 setLL += 13;
			 if (setL == 0){
			 debug(" \n\000");
			 for (i=0;i<setLL;i++){
			 if ((F_LName[i]==0xFF))  break; 
			 putchar1(F_LName[i]);
			 }
			 debug("\nF_StartCluster: %0\000",F_StartCluster);
			 debug("\nF_FileSize: %0\000",F_FileSize);
		 setLL = 0;
		 }
	   }
	}
	record++;
  }
 }
}

/******************************************************************************
**   Main Function  main()
******************************************************************************/
int main (void)
{
//  DWORD j;
  DWORD i;
  /* Fill block data pattern in write buffer and clear everything 
  in read buffer. */

  InitUART0();
  debug("UART Init ok!\n");
//  __asm("NOP"); 
  disk_initialize();
  /* on DMA channel 0, source is memory, destination is MCI FIFO. */
  /* On DMA channel 1, source is MCI FIFO, destination is memory. */
  FAT_init();
   debug("\nFAT init complite!\n");
  FindFiles();
  debug("---------------------------\n");
  /***************************************************************/
  /* TEST - Write and Read number of blocks defined by BLOCK_NUM */
  /***************************************************************/
//  debug("Test!\n");
//  for ( i = 0; i < 2; i++ )
//  {
////	debug("Step \n");
////	if ( MCI_Write_Block( i ) != TRUE )
////	{
////	 debug("Write error!\n");		/* Fatal error */
////	}
////	debug("Step1 \n");
////	/* When MCI_Block_End_Flag is clear, it indicates   
////	Write_Block is complete, next task, Read_Block to check write */
////   debug("Wait start!\n"); 
////	while ( MCI_Block_End_Flag == 1 );
////   debug("Wait complite!\n");
//	if ( MCI_Read_Block( i ) != TRUE )
//	{
//	  debug("Error MCI_Read_Block!\n");		/* Fatal error */
//	}
////	/* When MCI_Block_End_Flag is clear, it indicates RX is done 
////	with Read_Block,  validation of RX and TX buffers next. */
//	debug("Wait start!\n"); 
//	while ( MCI_Block_End_Flag == 1 );
//	debug("Wait complite!\n"); 	
//	for ( j = 0; j < (BLOCK_LENGTH); j++ )
//	{
//	   debug("\n %0\000",ReadBlock[j]); 	
//	}
////	/*  clear read buffer for next block comparison */
////	for ( j = 0; j < (BLOCK_LENGTH); j++ )
////	{
////	  ReadBlock[j] = 0;
////	}
//  }	
//  debug("Operation complite!\n"); 
//  return 0;
while (1) {;}
}

/******************************************************************************
**                            End Of File
******************************************************************************/
